/**
 * CHandler for JamochaMUD will act as our "proxy" between
 * the core of JamochaMUD and our different sockets, if we
 * have simultaneous connections to other MU*s
 * $Id: CHandler.java,v 1.10 2006/10/08 22:11:48 jeffnik Exp $
 */

/* JamochaMUD, a Muck/Mud client program
 * Copyright (C) 1998-2006 Jeff Robinson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */



package anecho.JamochaMUD;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;

import java.util.Vector;

import anecho.gui.JMText;
import anecho.gui.JMSwingText;
import anecho.JamochaMUD.legacy.MuckConnAWT;



/** To allow an easier time of dealing with multiple connections,
 * any input or output from JamochaMUD will pass through the CHandler
 * class.  This class will then determine which MU* the outbound
 * traffic will go out (eg. the &quot;active" MU* if from the DataIn
 * window) or the appropriate window to send incoming traffic
 * based on the socket it is received from.  CHandler now extends
 * the Container class so that it can act as it's own autonomous
 * "widget" coordinating views of MU*s and their actual connections.
 *
 * CHandler will &quot;hold" the actual MuSockets but for considering
 * which connection is visible/active, we will always ask our display
 * object, be it the JMTextPanel or the JTabbedPane.
 * @version $Id: CHandler.java,v 1.10 2006/10/08 22:11:48 jeffnik Exp $
 */

public class CHandler {
    
    
    
    final private Vector connections;   // A list of our active sockets/MUs
    final private JMConfig settings;          // a pointer to our settings
    final private boolean useSwing;   // Whether we are using Swing or not
    
    private static final boolean DEBUG = false;
    
    
    
    /**
     * Our constructor.  Since this should only be called once
     * per invocation of JamochaMUD, we will set up the &quot;frames"
     * for our MU* client.
     */
    
    // public CHandler(MuckMain mainWindow) {
    public CHandler() {
        
        settings = JMConfig.getInstance();
        
        connections = new Vector(0, 1);
        
        if (settings.getJMboolean(JMConfig.USESWING)) {
            useSwing = true;
        } else {
            useSwing = false;
        }
        
    }
    
    
    
    /**
     * Return the layout object that contains our MU* views
     * @return Returns a component that houses JamochaMUD's output display.
     */
    
    public Component getTextObject() {
        
        Component retComp = (Component)null;
        
        if (useSwing) {
            retComp = (anecho.gui.JMFancyTabbedPane)settings.getJMObject(JMConfig.TEXTPANEL);
        } else {
            retComp = (anecho.JamochaMUD.legacy.JMTabPanel)settings.getJMObject(JMConfig.TEXTPANEL);
        }
        
        return retComp;
    }
    
    
    
    /**
     * Sets the active MU* (the MU* visible to the user) based on the index
     * number provided.
     * @param target The index number repesenting which MU* should be set active.
     */
    
    public void setActiveMU(final int target) {
        
        if (DEBUG) {
            System.err.println("CHandler.setActiveMU: " + target);
        }
        
        if (target >= 0) {
            if (useSwing) {
                final anecho.gui.JMFancyTabbedPane tempPanel = (anecho.gui.JMFancyTabbedPane)getTextObject();
                
                try {
                    tempPanel.setSelectedIndex(target);  // Fix me!!
                } catch (Exception except) {
                    System.err.println("CHandler.setActiveMU exception: " + except);
                    return;
                }
                
//                tempPanel.invalidate();
//                tempPanel.validate();
//                tempPanel.doLayout();
                
            } else {
                final anecho.JamochaMUD.legacy.JMTabPanel tempPanel = (anecho.JamochaMUD.legacy.JMTabPanel)getTextObject();
                
                if (DEBUG) {
                    System.err.println("CHandler.setActiveMU() going to setSelectedIndex.");
                }
                
                tempPanel.setSelectedIndex(target);  // Fix me!
                
                if (DEBUG) {
                    System.err.println("CHandler.setActiveMU() invalidating.");
                }
                
                tempPanel.invalidate();
                
                if (DEBUG) {
                    System.err.println("CHandler.setActiveMU() validating.");
                }
                
                tempPanel.validate();
                
                if (DEBUG) {
                    System.err.println("CHandler.setActiveMU() doLayout.");
                }
                
                tempPanel.doLayout();
                
            }
            
            // Run through our list of MU*s and tell them whether they are
            // active or inactive  Fix Me XXX
            MuSocket tempMU;
            final int conns = connections.size();
            for (int i = 0; i < conns; i++) {
                tempMU = this.getMUHandle(i);
                if (i == target) {
                    tempMU.setActiveMU(true);
                    if (DEBUG) {
                        System.err.println(tempMU.getMUName() + " set to true");
                    }
                } else {
                    tempMU.setActiveMU(false);
                    if (DEBUG) {
                        System.err.println(tempMU.getMUName() + " set to false");
                    }
                }
            }
            
            if (DEBUG) {
                System.err.println("CHandler.setActiveMU() complete.");
            }
            
            
            
        }
        
    }
    
    
    
    /**
     * This method displays the last MU* connected to that is still open.
     */
    
    // public synchronized void showLatestMU() {
    public void showLatestMU() {
        
        if (connections.size() < 1) {
            // There are no windows left to show.
            System.out.println("CHandler: No (additional) connections left to show.");
            return;
        }
        
        if (DEBUG) {
            System.err.println("Chandler.showLatestMU()");
        }
        
        if (useSwing) {
            final anecho.gui.JMFancyTabbedPane tempPanel = (anecho.gui.JMFancyTabbedPane)getTextObject();
            // tempPanel.setSelectedIndex(tempPanel.getTabCount() - 1);
            this.setActiveMU(tempPanel.getTabCount() - 1);
            tempPanel.invalidate();
            tempPanel.validate();
            tempPanel.doLayout();
        } else {
            final anecho.JamochaMUD.legacy.JMTabPanel tempPanel = (anecho.JamochaMUD.legacy.JMTabPanel)getTextObject();
            tempPanel.last();
            tempPanel.invalidate();
            tempPanel.validate();
            tempPanel.doLayout();
        }
        
    }
    
    
    
    /**
     * Add another MU* to our display object
     * @param name The human-readable name of the MU*
     * @param address The numeric or human-readable address for the MU* server
     * @param port The MU* server port to connect to
     * @param msock The MuSocket which is responsible for communicating between
     * JamochaMUD the new MU*.
     */
    // public synchronized void addNewMU(final String name, final String address, final int port, final MuSocket msock) {
    public void addNewMU(final String name, final String address, final int port, final MuSocket msock) {
        if (DEBUG) {
            System.err.println("CHandler.addNewMU Adding " + name + " tab.");
        }
        
        if (useSwing) {
            
            final JMSwingText tempText = msock.getSwingTextWindow();
            final anecho.gui.JMFancyTabbedPane tempPane = (anecho.gui.JMFancyTabbedPane)getTextObject();
            
            tempPane.addTab(name, tempText);
            
            
        } else {
            final JMText tempText = msock.getTextWindow();
            
            ((anecho.JamochaMUD.legacy.JMTabPanel)getTextObject()).addTab(name, tempText);
            
        }
        
        if (DEBUG) {
            System.err.println("showLatestMU.");
        }
        
        showLatestMU();
    }
    
    /**
     * Create a new socket and add a &quot;window" for it
     * @return Returns a new MuSocket object that will be responsible for
     * this connection.
     * @param ssl Returns whether the connection is to be encrypted (SSL)
     * or plain text.
     * <CODE>true</CODE> - encrypted connection
     * <CODE>false</CODE> - plain text connection
     * @param name The human-readable name for the new MU*
     * @param address The human-readable or numeric address for the MU* server
     * @param port The MU* server port to connect to
     */
    // public synchronized void OpenSocket(String name, String address, int port) {
    // public synchronized MuSocket openSocket(final String name, final String address, final int port, final boolean ssl) {
    public MuSocket openSocket(final String name, final String address, final int port, final boolean ssl) {
        // create a new MuSocket, and then populate it with the proper info
        if (DEBUG) {
            System.err.println("CHandler.OpenSocket: " + name + " " + address + " " + port);
        }
        final MuSocket msock = new MuSocket();
        
        msock.setTitle(name);
        msock.setAddress(address);
        msock.setPort(port);
        msock.setSSL(ssl);
        
        // We'll set a timestamp to this MU* to differentiate incase the
        // user hooks up to the same MU* multiple times.
        msock.setTimeStamp(System.currentTimeMillis());
        
        // Set our logging option
        msock.setLogging(settings.getJMboolean(JMConfig.AUTOLOGGING));
        
        // Now add the new Socket to the connections vector
        connections.addElement(msock);
        
        // Now we'll start our socket-listening thread
        msock.start();
        
        // Update the connection menu on the main program?
        // Fix me XXX!
        
        // System.out.println("CHandler.openSocket returns: " + msock);
        return msock;
        
    }
    
    /** Make the next MU* in the list the active MU*
     *
     * The process that makes the call to this method
     * is responsible for updating the display appropriately
     *
     */
    public void nextMU() {
        
        if (useSwing) {
            
            final anecho.gui.JMFancyTabbedPane tempPanel = (anecho.gui.JMFancyTabbedPane)getTextObject();
            
            int muNum = tempPanel.getSelectedIndex();
            
            muNum++;
            
            if (muNum > (tempPanel.getTabCount() - 1)) {
                
                muNum = 0;
                
            }
            
            tempPanel.setSelectedIndex(muNum);
            
        } else {
            
            final anecho.JamochaMUD.legacy.JMTabPanel tempPanel = (anecho.JamochaMUD.legacy.JMTabPanel)getTextObject();
            
            tempPanel.next();
            
        }
        
    }
    
    
    
    /** Make the previous MU* in the list the active MU*.
     *
     * As with nextMU, the calling process is responsible
     *
     * for updating the display appropriately.
     *
     */
    
    public void previousMU() {
        
        
        
        if (useSwing) {
            
            final anecho.gui.JMFancyTabbedPane tempPanel = (anecho.gui.JMFancyTabbedPane)getTextObject();
            
            int muNum = tempPanel.getSelectedIndex();
            
            muNum--;
            
            if (muNum < 0) {
                
                muNum = tempPanel.getTabCount() - 1;
                
            }
            
            tempPanel.setSelectedIndex(muNum);
            
        } else {
            
            final anecho.JamochaMUD.legacy.JMTabPanel tempPanel = (anecho.JamochaMUD.legacy.JMTabPanel)getTextObject();
            
            tempPanel.previous();
            
        }
        
    }
    
    
    
    /** Close the connection to the active MU*
     *
     * @param stamp Timestamp of MU* to be closed (as confirmation)
     *
     */
    
    public void closeActiveMU(final String stamp) {
        
        final MuSocket active = getActiveMUHandle();
        
        closeSocket(active);
        
    }
    
    
    
    /**
     *
     * Remove the active MU* from our list altogether
     *
     * @param stamp This is the time stamp used to verify that we are closing
     *
     * the proper MU*.
     *
     */
    
    public void removeActiveMU(final String stamp) {
        
        // Disconnect the MU* if still connected
        
        if (isActiveMUDConnected()) {
            
            closeActiveMU(stamp);
            
        }
        
        
        
        // "destroy" any physical entities
        // uhm... somehow.  We could rely on garbage collection but
        // I'd feel better actually knowing it was gone.
        // Fix this XXX
        // remove the MU* from our hashtable
        
        final int aMU = getActiveMUIndex();
        
        connections.removeElementAt(aMU);
    }
    
    
    
    /**
     *
     * Remove the MU* indicated by the MuSocket and timeStamp
     *
     * @param mSock The MuSocket to close and remove
     *
     */
    
    public void removeMU(final MuSocket mSock) {
        
        if (mSock.isConnectionActive()) {
            
            if (DEBUG) {
                
                System.err.println("CHandler.removeMU(): Disconnecting active socket");
                
            }
            
            closeSocket(mSock);
            
        }
        
        
        
        int place;
        
        place = connections.indexOf(mSock);
        
        
        
        // Remove this from out layout as well
        
        if (useSwing) {
            
            final JMSwingText tempText = mSock.getSwingTextWindow();
            
            
            
            final int index = ((anecho.gui.JMFancyTabbedPane)getTextObject()).indexOfComponent(tempText);
            
            if (DEBUG) {
                
                System.err.println("CHandler.removeMU(): The index of our component is: " + index);
                
            }
            
            
            
            ((anecho.gui.JMFancyTabbedPane)getTextObject()).removeTabAt(index);
            
        } else {
            
            final JMText tempText = mSock.getTextWindow();
            
            ((anecho.JamochaMUD.legacy.JMTabPanel)getTextObject()).remove(tempText, place);
            
            
            
        }
        
        
        
        if (DEBUG) {
            
            System.err.println("CHandler.removeMU(): Component removed");
            
        }
        
        
        
        connections.removeElement(mSock);
        
        
        
        if (DEBUG) {
            
            System.err.println("CHandler.removeMU(): Completed successfully");
            
        }
        
    }
    
    
    
    /**
     *
     * Close and &quot;destroy" an existing socket.
     *
     * This allows any of our clases to shut things down
     *
     * @param socket The socket to be closed.
     *
     */
    
    public void closeSocket(final MuSocket socket) {
        
        
        
        // Clean up the existing socket
        
        socket.closeSocket();
        
        
        
        // Removed for testing purposes
        
        // Set the menu state
        
        // MuckMain main = settings.getMainWindowVariable();
        
        // main.setConnectionMenu();
        
    }
    
    
    
    /**
     *
     * Return the 'address' for the active JMText object
     *
     * @deprecated use JMText getActiveMUDText(String stamp)
     *
     * @return Returns a variable representing the active MU*
     *
     */
    
    public JMText getActiveMUDText() {
        
        final int aMU = getActiveMUIndex();
        
        final MuSocket msock = (MuSocket)connections.elementAt(aMU);
        
        return msock.getTextWindow();
        
    }
    
    
    
    /**
     *
     * This returns the active JMSwingText
     *
     * @return JMSwingText of the active MU*
     *
     */
    
    public JMSwingText getActiveMUDSwingText() {
        
        JMSwingText retText;
        
        
        
        if (connections.size() > 0) {
            
            final int aMU = getActiveMUIndex();
            
            final MuSocket msock = (MuSocket)connections.elementAt(aMU);
            
            // return msock.getSwingTextWindow();
            
            retText = msock.getSwingTextWindow();
            
        } else {
            
            retText = (JMSwingText)null;
            
        }
        
        
        
        return retText;
        
    }
    
    
    
    /**
     *
     * Return the active JMText component, verified with the passed timestamp
     *
     * @param stamp A timestamp to verify the component returned
     *
     * @return Returns a JMText component of the active MU*
     *
     */
    
    public JMText getActiveMUDText(final String stamp) {
        
        final int aMU = getActiveMUIndex();
        
        final MuSocket msock = (MuSocket)connections.elementAt(aMU);
        
        return msock.getTextWindow();
        
        
        
    }
    
    
    
    /**
     *
     * Return the index of the active MU
     *
     * @return Returns the index number of the active MU*
     *
     */
    
    public int getActiveMUIndex() {
        
        int index = 0;
        
        
        
        if (useSwing) {
            
            index = ((anecho.gui.JMFancyTabbedPane)getTextObject()).getSelectedIndex();
            
        } else {
            
            index = ((anecho.JamochaMUD.legacy.JMTabPanel)getTextObject()).getSelectedIndex();
            
        }
        
        
        
        return index;
        
    }
    
    
    
    /**
     *
     * Return the title of the MU based on the given index
     *
     * @param index Index representing the MU* we are to retrieve the title of.
     *
     * @return The title of the MU* represented by the given {@link index}
     *
     */
    
    public String getTitle(final int index) {
        
        String title;
        
        
        
        if (index < 0 || index > (connections.size() - 1)) {
            
            if (DEBUG) {
                
                System.err.println("CHandler.getTitle out of range.");
                
            }
            
            title = "";
            
        } else {
            
            title = ((MuSocket)connections.elementAt(index)).getTitle();
            
        }
        
        return title;
        
    }
    
    
    
    /**
     *
     * Return the proper MuSocket for the requested mu
     *
     * @param muNum This variable represents the index of the MU* to retrieve
     *
     * @return MuSocket representing the MU* specified by our {@link muNum}
     *
     */
    
    // public synchronized MuSocket getMUHandle(final int muNum) {
    public MuSocket getMUHandle(final int muNum) {
        
        MuSocket retSock;
        
        if (muNum < connections.size() && muNum > -1) {
            
            retSock = (MuSocket)connections.elementAt(muNum);
            
        } else {
            
            retSock = (MuSocket)null;
            
        }
        
        
        
        return retSock;
        
    }
    
    
    
    /**
     *
     * Return the MuSocket for the currently &quot;active" MU*
     *
     * @return The MuSocket of the currently active MU*
     *
     */
    
    public synchronized MuSocket getActiveMUHandle() {
        
        MuSocket retSock;
        
        
        
        // Check to see if we have any connections yet
        
        if (connections.size() > 0) {
            
            int aMU = 0;
            
            if (useSwing) {
                
                aMU = ((anecho.gui.JMFancyTabbedPane)getTextObject()).getSelectedIndex();
                
            } else {
                
                aMU = ((anecho.JamochaMUD.legacy.JMTabPanel)getTextObject()).getSelectedIndex();
                
            }
            
            
            
            retSock = (MuSocket)connections.elementAt(aMU);
            
        } else {
            
            retSock = (MuSocket)null;
            
        }
        
        
        
        return retSock;
        
    }
    
    
    
    /**
     *
     * Return the connection status of the active MU*
     *
     * @return <CODE>true</CODE> - The active MU* is connected to a server
     *
     * <CODE>false</CODE> - The active MU* is not connected to a server
     *
     */
    
    // public synchronized boolean isActiveMUDConnected() {
    
    public boolean isActiveMUDConnected() {
        
        boolean active = false;
        
        
        
        int aMU = 0;
        
        
        
        if (useSwing) {
            
            aMU = ((anecho.gui.JMFancyTabbedPane)getTextObject()).getSelectedIndex();
            
        } else {
            
            aMU = ((anecho.JamochaMUD.legacy.JMTabPanel)getTextObject()).getSelectedIndex();
            
        }
        
        
        
        if (aMU > connections.size() - 1) {
            
            if (DEBUG) {
                
                System.err.println("CHandler.isActiveMUConnected()");
                
                System.err.println("aMU: " + aMU + " and our connection size: " + connections.size());
                
            }
            
        } else {
            
            if (aMU >= 0 && aMU < connections.size()) {
                
                final MuSocket msock = (MuSocket)connections.elementAt(aMU);
                
                active = msock.isConnectionActive();
                
            }
            
        }
        
        
        
        return active;
        
    }
    
    
    
    /**
     *
     * Send a string to the active MU*, using the
     *
     * proper encoding method (ASCII or Unicode)
     *
     * @param send The String to be sent to the active MU*
     *
     */
    
    public void sendText(final String send) {
        
        int aMU = 0;
        
        
        
        if (useSwing) {
            aMU = ((anecho.gui.JMFancyTabbedPane)getTextObject()).getSelectedIndex();
        } else {
            aMU = ((anecho.JamochaMUD.legacy.JMTabPanel)getTextObject()).getSelectedIndex();
        }
        
        
        
        if (aMU > -1 && aMU < connections.size()) {
            
            final MuSocket msock = (MuSocket)connections.elementAt(aMU);
            
            
            
            msock.sendText(send);
            
        } else {
            
            if (DEBUG) {
                
                System.err.println("CHandler cannot send text as aMU " + aMU + " is not within the size of our connections " + connections.size());
                
            }
            
        }
        
    }
    
    
    
    /**
     *
     * Change the font face and size on all the connections
     *
     * @param newPal This array represents the colour palette that is to be used
     *
     * @param fontFace The Font to use for the display.
     *
     * @param fgColour The foreground colour to use
     *
     * @param bgColour The background colour to use
     *
     */
   
    public void setAllAttribs(final Font fontFace, final Color fgColour, final Color bgColour, final Color[] newPal) {
       MuSocket msock;       // our temporary socket
        JMText text;          // our temporary text object
        JMSwingText sText;
        
        final int total = connections.size();
        
        if (total < 1) {
            // There are no active MU*s
            return;
        }
        
        for (int i = 0; i < total; i++) {
            // Loop through our connections and change the fonts
            msock = (MuSocket)connections.elementAt(i);
            
            if (useSwing) {
                sText = msock.getSwingTextWindow();
                if (fontFace != null) {
                    sText.setFont(fontFace);
                    
                }
                
                if (fgColour != null) {
                    
                    sText.setForeground(fgColour);
                    
                    sText.setBackground(bgColour);
                    
                }
                
                if (newPal != null) {
                    
                    try {
                        
                        sText.setPalette(newPal);
                        
                    } catch (Exception exc) {
                        
                        if (DEBUG) {
                            
                            System.err.println("Exception settings new custom palette from CHandler.");
                            
                        }
                        
                    }
                    
                }
                
            } else {
                
                text = msock.getTextWindow();
                
                if (fontFace != null) {
                    
                    text.setFont(fontFace);
                    
                }
                
                if (fgColour != null) {
                    
                    text.setForeground(fgColour);
                    
                    text.setBackground(bgColour);
                    
                }
                
            }
            
        }
        
    }
    
    
    
    /**
     *
     * Activate or deactivate logging in all connections
     *
     * @param state <CODE>true</CODE> - enable logging
     *
     * <CODE>false</CODE> - disable logging
     *
     */
    
    public void setLogging(final boolean state) {
        
        MuSocket msock;       // our temporary socket
        
        final int total = connections.size();
        
        
        
        if (total < 1) {
            
            // There are no active MU*s
            
            return;
            
        }
        
        
        
        for (int i = 0; i < total; i++) {
            
            // Loop through our connections and change the fonts
            
            msock = (MuSocket)connections.elementAt(i);
            
            msock.setLogging(state);
            
        }
        
    }
    
    
    
    /**
     *
     * Set the foreground and background colours for all the MU*s.
     *
     * @param fgColour The foreground colour to use
     *
     * @param bgColour The background colour to use
     *
     */
    
    public void setAllColours(final Color fgColour, final Color bgColour) {
        
        setAllAttribs(null, fgColour, bgColour, null);
        
    }
    
    
    
    /**
     *
     * This method sets the custom colours to be used on all our MU* windows.
     *
     * @param newPal An array containing our new Colors.
     *
     */
    
    public void setCustomPalette(final Color[] newPal) {
        
        setAllAttribs(null, null, null, newPal);
        
    }
    
    
    
    /**
     *
     * Set the Font for all active MU*s
     *
     * @param newStyle The Font style to use
     *
     */
    
    public void setAllFonts(final Font newStyle) {
        
        setAllAttribs(newStyle, null, null, null);
        
    }
    
    
    
    /** Set-up a connection to a new MU* */
    
    public synchronized void connectToNewMU() {
        
        // Show our list of connections (the MuckConnector)
        if (DEBUG) {
            System.err.println("CHandler.connectToNewMU starting.");
        }
        
        // Check first to make certain we don't already have a connection dialogue
        if (!settings.getJMboolean(JMConfig.MUCKCONNVISIBLE)) {
            if (DEBUG) {
                System.err.println("CHandler.connectToNewMU creating new MuckConn.");
            }
            
            String name, address;
            int port;
            boolean ssl;
            String connStr = "unset";
            
            if (settings.getJMboolean(JMConfig.USESWING)) {
                // Try using our new World Connector
//                final MuckConn muChoice = new MuckConn(settings);
//                
                final AddEditWorld muChoice = new AddEditWorld();
                muChoice.setVisible(true);
                
                // Gather information about our chosen connection
                name = muChoice.getName();
                address = muChoice.getAddress();
                port = muChoice.getPort();
                ssl = muChoice.getSSL();

                connStr = muChoice.getConnectionString();
                
            } else {
                final MuckConnAWT muChoice = new MuckConnAWT(settings);
                
                muChoice.setVisible(true);
                
                // Gather information about our chosen connection
                
                name = muChoice.getName();
                address = muChoice.getAddress();
                port = muChoice.getPort();
                ssl = false;  // SSL is not part of Java 1.1
                // ssl = muChoice.getSSL();
                // Open our new connection
                connStr = null;
            }
            
            
            
            if (DEBUG) {
                
                System.err.println("CHandler.connectToNewMU calling OpenSocket for: " + name + " " + address + " " + port);
                
                
            }
            
            
            
            // Open a new socket if a MU* has been selected
            
            if (address != null && !(address.trim()).equals("")) {
                if (DEBUG) {
                    System.err.println("CHandler.connectToNewMU using name: " + name);
                }
                
                // final MuSocket msock = openSocket(name, address, port);
                final MuSocket msock = openSocket(name, address, port, ssl);
                if (connStr != null) {
                    msock.setAutoConnect(true);
                    msock.setConnectString(connStr);
                    // msock.setConnChar("help"); 
                } else {
                    msock.setAutoConnect(false);
                }
                
                if (DEBUG) {
                    System.err.println("CHandler.connectToNewMU calling addNewMU");
                }
                
                addNewMU(name, address, port, msock);

            } else {
                
                if (DEBUG) {
                    System.err.println("CHandler.connectToNewMU didn't have a MU* to connect to.");
                }
                
            }
            
            
        }
        
    }
    
    
    
    /**
     *
     * Get the total number of MU*s open regardless of their connection status
     *
     * @return The number of existing MuSockets
     *
     */
    
    // public synchronized int totalConnections() {
    public int totalConnections() {
        
        if (DEBUG) {
            System.err.println("CHandler.totalConnections(): entering method");
        }
        
        int retSize = 0;
        
        
        
        if (DEBUG) {
            
            System.err.println("CHandler.totalConnections() getting connection size.");
            
        }
        
        
        
        if (connections.size() > 0) {
            
            retSize = connections.size();
            
        }
        

        // public int totalConnections() {
        
        if (DEBUG) {
            
            System.err.println("CHandler.totalConnections() returns: " + retSize);
            
        }
        
        
        
        // return connections.size();
        if (DEBUG) {
            System.err.println("CHandler.totalConnections(): leaving method");
        }
        
        return retSize;
        
    }
    
    
    
    /**
     *
     * Determines if the given MU* is the currently active MU*.
     *
     * @return <CODE>true</CODE> - the provided MU* is the active MU*
     *
     * <CODE>false</CODE> - the provided MU* is not the active MU*
     *
     * @param mSock The MuSocket to check to see if it is current the active MU*.
     *
     */
    
    public boolean isActiveMU(final MuSocket mSock) {
        
        boolean active = false;
        
        
        
        // System.out.println("CHandler.isActiveMU...");
        
        // if (getActiveMUHandle() == mSock) {
        
        if (getActiveMUHandle().equals(mSock)) {
            
            active = true;
            
        }
        
        
        
        return active;
        
    }
    
    
    
    
    
    /**
     *
     * Return the title of the active MU*
     *
     * @return Return a String of the title of the active MU*.
     *
     */
    
    public synchronized String getActiveTitle() {
        
        String title;
        
        
        
        if (connections.size() < 1) {
            
            // We have no connections, thus no title
            
            title = "";
            
        } else {
            
            // MuSocket msock = (MuSocket)connections.elementAt(activeMU);
            
            // title = msock.getTitle();
            
            final int aMU = getActiveMUIndex();
            
            title = getTitle(aMU);
            
        }
        
        if (DEBUG) {
            
            System.out.println("getActiveTitle returns: " + title);
            
        }
        
        return title;
        
    }
    
    
    
    /**
     *
     * Change the drawing type for the JMText... either
     *
     * single buffered (false) or double buffered (true)
     *
     * @param state <CODE>true</CODE> - set the display to double buffer
     *
     * <CODE>false</CODE> - set the display to work without buffering
     *
     */
    
    public void setDoubleBuffer(final boolean state) {
        
        MuSocket msock;
        
        for (int i = 0; i < connections.size(); i++) {
            
            msock = (MuSocket)connections.elementAt(i);
            
            (msock.getTextWindow()).setDoubleBuffer(state);
            
        }
        
    }
    
    
    
    /**
     *
     * Change between aliased and antialised text in our
     *
     * output windows.
     *
     * @param state <CODE>true</CODE> - enable antialiasing
     *
     * <CODE>false</CODE> - disalbe antialiasing
     *
     */
    
    public void setAntiAliasing(final boolean state) {
        
        MuSocket msock;
        
        for (int i = 0; i < connections.size(); i++) {
            
            if (DEBUG) {
                
                System.out.println("CHandler.setAntiAliasing setting connection " + i + " to " + state);
                
            }
            
            msock = (MuSocket)connections.elementAt(i);
            
            (msock.getSwingTextWindow()).setAntiAliasing(state);
            
        }
        
    }
    
    
    
    /**
     *
     * Set the MU* displays to operating in either "low colour" (8 colours,
     *
     * normal and bold) or "high colour" (16 colours).
     *
     * @param state <CODE>true</CODE> - set the display to low colour mode
     *
     * <CODE>false</CODE> - set display to high colour mode
     *
     */
    
    public void setLowColour(final boolean state) {
        
        MuSocket msock;
        
        for (int i = 0; i < connections.size(); i++) {
            
            if (DEBUG) {
                
                System.out.println("CHandler.setLowColour setting connection " + i + " to " + state);
                
            }
            
            msock = (MuSocket)connections.elementAt(i);
            
            (msock.getSwingTextWindow()).setBoldNotBright(state);
            
            
            
        }
        
    }
    
    
    
    /**
     *
     * Return the ECHO state for the active MU*
     *
     * @return <CODE>true</CODE> - the active MU* echo mode is enabled
     *
     * <CODE>false</CODE> - the active MU* is not in echo mode
     *
     */
    
    public boolean getActiveMUEchoState() {
        
        final MuSocket active = getActiveMUHandle();
        
        return active.getEchoState();
        
    }
    
    
    
    /** Reconnect to the currently active (visible) MU* */
    
    public synchronized void reconnectToMU() {
        
        
        
        final MuSocket handle = getActiveMUHandle();
        
        
        
        if (handle == null) {
            
            // No MU*s have previously been connected to, so we can't reconnect
            
            return;
            
        }
        
        
        
        final String name = handle.getMUName();
        
        final String address = handle.getAddress();
        
        final int port = handle.getPort();
        
        final boolean ssl = handle.getSSL();
        
        
        
        if (DEBUG) {
            
            System.err.println("Chandler.reconnectToMU activeMU Handle is: " + handle);
            
            System.err.println("CHandler.reconnectToMU attempting connection to: " + name + " " + address + " " + port);
            
            System.err.println("CHandler.reconnectToMU calling OpenSocket");
            
        }
        
        
        
        // final MuSocket msock = openSocket(name, address, port);
        
        final MuSocket msock = openSocket(name, address, port, ssl);
        
        
        
        if (DEBUG) {
            
            System.err.println("CHandler.reconnectToMU Adding the new MU*");
            
        }
        
        
        
        addNewMU(name, address, port, msock);
        
        
        
        if (DEBUG) {
            
            System.err.println("CHandler.reconnectToMU removing old MU*.");
            
        }
        
        
        
        removeMU(handle);
        
        
        
        if (DEBUG) {
            
            System.err.println("CHandler.reconnectToMU completed.");
            
        }
        
        
        
    }
    
}

